/* File: boggle.cpp
 * ----------------
 * Author:  Yinyi Guo
 * Date: 02/15
 * Section leader: Ben Phillips
 */

#include "genlib.h"
#include "gboggle.h"
#include "sound.h"
#include "extgraph.h"
#include "simpio.h"
#include <iostream>
#include <fstream>
#include <string>
#include "grid.h"
#include "vector.h"
#include "random.h"
#include "lexicon.h"


struct block {       // block stores the letter on it
	char letter;     // a boolean, if it's tried before
	bool path;	
};

static const string dice = "dice rattle.wav";
static const string denied = "denied.wav";
static const string out = "moo.wav";
static const string illegal = "not fooling anyone.wav";
static const string defeat = "oh really.wav";
static const string again = "thats pathetic.wav";
static const string findone = "tinkerbell.wav";
static const string notenough = "whoops.wav";
static const string gosh = "tweetle.wav";
static const string emm = "yeah right.wav";

/*
 * initBoard() refresh the every path of block on the board to false 
 */
void initBoard(Grid<block>& b)          
{						
	for(int i=0;i< b.numRows();i++) {
		for (int j=0; j< b.numCols();j++) {
			b(i,j).path=false;
		}
	}
}

/*
 *  isOnBoard() return if this position is on the board 
 */
bool isOnBoard(int row,int col)  
{
	if (row>3||row<0||col>3||col<0)
		return false;
	return true;
}

/*
 *  exhausted() is for computer to prints out all the legal words human left 
 */
void exhausted(Lexicon& lex, Grid<block>& b,string t, int row,int col, Lexicon& cache, int& s)
{                                     
	if ((!lex.containsPrefix(t))) return;  
	if (lex.containsWord(t)&&(!cache.containsWord(t))&&t.size()>3) {  // base case 
		cache.add(t);
		s+= t.size()-3;
		RecordWordForPlayer(t,Computer);
	}
	if ((isOnBoard(row,col))&&(b(row,col).path==false)) {  // on board and not tried
		char p = b(row,col).letter;
		b(row,col).path= true;
		string temp= t;
		exhausted(lex,b,t+p,row-1,col+1,cache,s);     // tried each neighbor
		exhausted(lex,b,t+p,row,col+1,cache,s); 
		exhausted(lex,b,t+p,row+1,col+1,cache,s);
		exhausted(lex,b,t+p,row-1,col,cache,s); 
		exhausted(lex,b,t+p,row+1,col,cache,s);
		exhausted(lex,b,t+p,row-1,col-1,cache,s); 
		exhausted(lex,b,t+p,row,col-1,cache,s); 
		exhausted(lex,b,t+p,row+1,col-1,cache,s);
		b(row,col).path = false;                      // trigger back tracking
		t = temp;
	}
}

/*
 *  return the score when the computer finish playing, using exhausted recursion  
 */ 
int computerTurn(Grid<block>& b, Lexicon& lex, Lexicon& cache)
{
	int score =0;
	for (int row=0;row<b.numRows();row++) {         // try each starting point
		for (int col=0;col<b.numCols();col++) {
			exhausted(lex, b,"",row,col,cache, score);  // recursion for computer to exhausted the remaining 
		}
	}
	return score;
}

/*
 * this is recursive helper function to return if it's legal for human to find that input word  
 */ 
bool isLegalH(Lexicon& lex, Grid<block>& b, string word, string t,int row, int col,Lexicon& cache)
{                                       
	if (t== word) {    // base case
		cache.add(word);
		return true;
	}
	if (!lex.containsPrefix(t)) return false;
	if ((isOnBoard(row,col))&&(b(row,col).path==false)) {
		char p = b(row,col).letter;
		b(row,col).path= true;
		string temp= t;
		if (isLegalH(lex,b,word,t+p,row-1,col+1,cache))  return true;  // tried each neighbor
		if (isLegalH(lex,b,word,t+p,row,col+1,cache))  return true;
		if (isLegalH(lex,b,word,t+p,row+1,col+1,cache))  return true;
		if (isLegalH(lex,b,word,t+p,row-1,col,cache))  return true;
		if (isLegalH(lex,b,word,t+p,row+1,col,cache))  return true;
		if (isLegalH(lex,b,word,t+p,row-1,col-1,cache))  return true;
		if (isLegalH(lex,b,word,t+p,row,col-1,cache))  return true;
		if (isLegalH(lex,b,word,t+p,row+1,col-1,cache))  return true;
		b(row,col).path = false;           // trigger the back tracking
		t = temp;
	}
	return false;
}

/*
 * this is function to return if it's legal for human to find that input word, using one helper function 
 */ 
bool isLegal(Lexicon& lex, Grid<block>& b,string word, string t,Lexicon& cache)
{
	for (int row=0;row<b.numRows();row++) {   // try every starting point
		for (int col=0;col<b.numCols();col++) {
			if (b(row,col).letter==word[0]) {
				if (isLegalH(lex,b,word,"",row,col,cache))
				return true;
			}
		}
	}
	return false; 
}

/*
 * this is function to return the score when the player finish playing
 */ 
int userTurn(Grid<block>& b, Lexicon& lex, Lexicon& cache, bool sound)
{
	int score =0;
	while(true) {
		cout<<"Enter a word: ";
		initBoard(b);
		string word =GetLine();
		word = ConvertToUpperCase(word);
		if (word.size()< 1) return score; // input a blank line to switch to pc
		if (word.size()<4) {   // word length is too short!
			if (sound) {
				PlayNamedSound(notenough);
			}
			cout<<"I'm sorry, but we have our standards."<<endl;
			cout<<"That word doesn't meet the minimum word length."<<endl;
		}
		else if (!lex.containsWord(word)) {  // that's not a word in lexicon
			if (sound) {
				PlayNamedSound(illegal);
			}
			cout<<"But that's not a word!"<<endl;
		}
		else if (cache.containsWord(word)) { // you have guessed that before!
			if (sound) {
				PlayNamedSound(again);
			}
			cout<< "You've already guessed that!"<<endl;
		}
		else if (!isLegal(lex, b, word,"",cache)) { // not legal on the board(detect by recursion) 
			if (sound) {
				PlayNamedSound(denied);
			}
			cout<<"You can't make that word!"<<endl;
		}
		else {         // finally, you get one
			if (sound) {
				PlayNamedSound(findone);
			}
			for(int i=0;i< b.numRows();i++) {     // highlight the word for a while
				for (int j=0; j< b.numCols();j++) {
					HighlightCube(i,j,b(i,j).path);
				}
			}
			initBoard(b);                    // and stop highlighting
			Pause(0.3);
			for(int i=0;i< b.numRows();i++) {
				for (int j=0; j< b.numCols();j++) {
					HighlightCube(i,j,b(i,j).path);
				}
			}
			score += word.size()-3;      // compute the score
			word =ConvertToLowerCase(word);  // show up the word in lower case
			RecordWordForPlayer(word,Human); // show record and words
		}
	}
}

/*
 * this is function to randomly arrange the board by the offered dice
 */ 
void DiceShaking(Grid<block>& b, bool sound)
{
	Randomize();
	ifstream infile;
	infile.open("cubes25.txt");   // not 4x4!~
	string next,temp;
	Vector<string> vec;
	while(true) {                 // DiceReading
		if (infile.fail()) break;
		getline(infile,next,'\r');
		vec.add(next);
	}
	for (int i=0;i<vec.size()-1;i++) {        // random choice one
		b(i/5,i%5).letter=vec[i][RandomInteger(0,vec[i].size()-1)];
	}
	for (int row = 0;row<b.numRows();row++) {  // random choice two
		for (int col=0;col<b.numCols();col++) {
			char temp = b(row,col).letter;
			int rowNew = RandomInteger(row, b.numRows()-1);
			int colNew = RandomInteger(col, b.numCols()-1);
			b(row,col).letter= b(rowNew,colNew).letter;
			b(rowNew,colNew).letter=temp;
		}
	}
	if (sound) {
		PlayNamedSound(dice);
	}
	for (int i=0;i<b.numRows();i++) {      // label each cube on the board
		for (int j=0;j<b.numCols();j++) {
			LabelCube(i,j,b(i,j).letter);
			b(i,j).path = false;           // init each path back to false 
		}
	}
	cout<<"Ok, take all the time you want and find all the words you can!"<<endl;
	cout<<"Signal that you're finished by entering an empty line."<<endl;
	return;	
}

/*
 * this is function to force the board by inputing the letter by human hand
 */ 
void ForceTheBoard(Grid<block>& b)
{
	cout<<"Enter a 25-character string to identify which letters you want on the cubes."<<endl;
	cout<<"The first 5 letters are the cubes on the top row from left to right"<<endl;
	cout<<"next 5 letters are the second row, etc."<<endl;
	cout<<"Enter the string:";
	while(true) {
		string boo = GetLine();
		boo =ConvertToUpperCase(boo);
		if (boo.size()>=25) {           // only need 25 letters
			boo = boo.substr(0,25);
			for (int i=0;i< 25;i++) {  // arrange them onto the board
				b(i/5,i%5).letter =boo[i];
			}
			cout<<"Ok, take all the time you want and find all the words you can!"<<endl;
			cout<<"Signal that you're finished by entering an empty line."<<endl;
			for (int i=0;i<b.numRows();i++) {    // and show up
				for (int j=0;j<b.numCols();j++) {
					LabelCube(i,j,b(i,j).letter);
				}
			}
			return;
		}
		else cout<<"String must include 25 characters! Try again: ";
	}
}

/*
 * this is function for user to choose if they like to config the board by hand or randomly
 */ 
void Configuration(Grid<block>& b, bool sound)
{
	cout<< endl<< "I'll give you a chance to set up the board to your specification."<<endl
	<<"This makes it easier to confirm your boggle program is working."<<endl;
	while (true) {
		cout<<"Do you want to force the board configuration? ";
		string boo = GetLine();
		if (boo[0]=='y') {
			ForceTheBoard(b);
			return;
		}
		else if (boo[0]=='n') {
			DiceShaking(b,sound);
			return;
		}
		else cout<<"Please answer yes or no."<<endl;
	}
}

/*
 * this is function for user to choose if they like the instruction for the game
 */ 
void GiveInstructions()
{
	while (true) {
		cout << "Do you need instruction?";
		string boo = GetLine();
		if (boo[0]=='y') {
			cout << endl << "The boggle board is a grid onto which I will randomly distribute" << endl
			<< "dice.  These 6-sided dice have letters rather than numbers on the faces, " << endl
			<< "creating a grid of letters on which you try to form words.  You go first, " << endl
			<< "entering the words you find that are formed by tracing adjoining " << endl
			<< "letters.  Two letters adjoin if they are next to each other horizontally, " << endl
			<< "vertically, or diagonally. A letter can only be used once in the word. Words" << endl
			<< "must be at least 4 letters long and can only be counted once.  You score points" << endl
			<< "based on word length, a 4-letter word is worth one, 5-letters two, etc.  After your " << endl
			<< "tiny brain is exhausted, I, the brilliant computer, will find all the remaining " << endl
			<< "words in the puzzle and double or triple your paltry score." << endl;
			cout << "\nHit return when you're ready...";
			GetLine();
			return;
		}
		else if (boo[0]=='n')
			return;
		else cout<<"Please answer yes or no."<<endl;
	}	
}

/*
 * this is function for user to choose if they like sound effect for the game
 */
bool Sound()
{
	while (true) {
		cout << "Turn on the speaker for sound effect?";
		string boo = GetLine();
		if (boo[0]=='y') {
			cout<<"Music!"<<endl;
			return true;
		}
		else if (boo[0]=='n') {
			cout<<"why not..."<<endl;
			return false;
		}
		else cout<<"Please answer yes or no."<<endl;
	}
}

/*
 * welcome 
 */
void Welcome()
{
    cout << "Welcome!  You're about to play an intense game of mind-numbing Boggle." << endl
	 << "The good news is that you might improve your vocabulary a bit.  The" << endl
	 << "bad news is that you're probably going to lose miserably to this little" << endl
	 << "dictionary-toting hunk of silicon.  If only YOU had a gig of RAM...\n" << endl;
}

/*
 * entry point
 */
int main()
{
	Lexicon lex("lexicon.dat");
	Grid<block> b(5,5);
	InitGraphics();
	Welcome();
	bool sound = Sound();
	GiveInstructions();
	while (true) {
		Lexicon cache;
		InitGraphics();
		DrawBoard(5,5);
		Configuration(b,sound);
		int hm =userTurn(b,lex,cache,sound); 
		if (sound)
			PlayNamedSound(gosh);
		int pc =computerTurn(b,lex,cache);
		
		cout<<hm<<"vs"<<pc<<endl;
		if (sound&&(hm>=pc))  // you would hear that if you defeated the pc
			PlayNamedSound(defeat);
		else if (sound) PlayNamedSound(emm);
		while (true) {
			cout<<"Would you like to play again?";
			string boo = GetLine();
			if (boo[0]=='y') {
				break;
			}
			else if (boo[0]=='n') {
				cout<<"More practice, dude!"<<endl;
				if (sound) {
					PlayNamedSound(out);
				}
				return 0;
			}
			else cout<<"Please answer yes or no."<<endl;
		}
	}
}